/**
 *	@file	tiny_gettext.c
 *	@brief	A substitute of gettext.
 *	@author tenk* (Masashi Kitamura)
 *	@note
 *	- It is mainly for Windows.
 *	- Read po file (not mgo).
 *	- Win32: Read from .exe's directory.
 *	  Other: $HOME directory.
 *	- Win32: .po file name is 'exename.LCID.po',
 *			LCID: return value of Win32-API's GetThreadLocale().
 *				 ex) Japanese LCID is 1041, "exename.1041.po"
 *	  Other: ".domainname.LANG.po"
 *			LANG: return value of getenv("LANG").
 *				 ex) ".domainname.ja_JP.UTF-8.po"
 *	-	Public Domain Software
 */

#include <stdio.h>
#include <assert.h>
#include <sys/stat.h>
#include "tiny_gettext.h"


#ifdef _WIN32
 #include <windows.h>
 #define IsMbbLead(c)			IsDBCSLeadByte(c)
#else
 #if defined(_MSC_VER) || defined(__MINGW32__) || defined(__DMC__) || defined(__BORLANDC__) || defined(__WATCOMC__)
  #define HAVE_MBSTRING_H	1
 #endif
 #if HAVE_MBSTRING_H
  #include <mbstring.h>
  #define IsMbbLead(c)			_ismbblead(c)
 #else	/* no support */
  #define IsMbbLead(c)			(0)
 #endif
#endif



/* -------------------------------------------------------------------------*/

#ifndef TIGTTX_FILE_SIZE_MAX
#define	TIGTTX_FILE_SIZE_MAX		0x400000
#endif
#ifndef TIGTTX_MAX_PATH
#define TIGTTX_MAX_PATH				0x8001
#endif

#define STDERR						stderr

#ifndef FREE
#define CALLOC(n,sz)				calloc(n,sz)
#define REALLOC(p,sz)				realloc(p,sz)
#define FREE(p)						free(p)
#endif


/* -------------------------------------------------------------------------*/

struct TiGtTx_MsgPair {
	char*	id_;
	char*	str_;
};
typedef struct TiGtTx_MsgPair	TiGtTx_MsgPair;

static unsigned			tiGtTx_msgConv_num;
static TiGtTx_MsgPair*	tiGtTx_msgConvTbl;

static const char*		tiGtTx_domainName;
static char*			tiGtTx_workBuf;
static char*			tiGtTx_file_name;
static unsigned			tiGtTx_file_line;



/* -------------------------------------------------------------------------*/

#define TIGTTX_ERRMSG(x)	fprintf  x


#if 0 //def __cplusplus
extern "C" const char * gettext(const char *  msgId);
extern "C" void 		textdomain(const char *domainName);
extern "C" const char *	bindtextdomain(const char* domainName, const char* dir);
#endif


static int	 		tiGtTx_msgConvTbl_search(TiGtTx_MsgPair tbl[], int nn, const char* key);
static int	 		tiGtTx_msgConvTbl_insert(TiGtTx_MsgPair tbl[], int* tblcnt, const TiGtTx_MsgPair* key);
static const char* 	tiGtTx_skipspc(const char* s);
static void  		tiGtTx_strEsc(char** ppDst, char const* * ppSrc);
static int	 		tiGtTx_getMsgPair(TiGtTx_MsgPair* pMsgPair, char const*  * ppStr);
static char* 		tiGtTx_fileLoad(const char* fname);
static char* 		tiGtTx_getBasePath(void);
static char* 		tiGtTx_getLocaleStr(void);
static char* 		tiGtTx_getMsgFileName(void);
static char*		tiGtTx_getStr(char const* * ppStr);




/* -------------------------------------------------------------------------*/

const char *  gettext(const char *	msgId)
{
	if (tiGtTx_msgConv_num > 0) {
		int  n;
		n = tiGtTx_msgConvTbl_search(tiGtTx_msgConvTbl, tiGtTx_msgConv_num, msgId);
		if (n >= 0)
			return 	tiGtTx_msgConvTbl[n].str_;
	}
	return (const char *)msgId;
}



/* * It has not corresponded to the plural form. */
const char *  ngettext(const char *  msgId, const char * msgId2, int n)
{
  #if 1 /* fake */
	const char *  p = gettext(msgId);
	if (p == msgId && n != 1)
		p = msgId2;
	return p;
  #else /* fake */
	return	gettext((n == 1) ? msgId : msgId2);
  #endif
}



const char* textdomain(const char * domainName )
{
	char*	msgFileName 	= tiGtTx_getMsgFileName();
	char*	msgBuf			= tiGtTx_fileLoad(msgFileName);
	tiGtTx_domainName		= domainName;
	if (msgBuf) {
		unsigned 	size	= 0x80;
		tiGtTx_msgConv_num 	= 0;
		tiGtTx_msgConvTbl	= (TiGtTx_MsgPair*)CALLOC(1, sizeof(TiGtTx_MsgPair) * size );
		if (tiGtTx_msgConvTbl) {
			tiGtTx_workBuf	= (char*)CALLOC( 1, strlen(msgBuf)+16 );
			if (tiGtTx_workBuf) {
				const char* s	 = msgBuf;
				tiGtTx_file_name = msgFileName;
				tiGtTx_file_line = 1;
				for (;;) {
					TiGtTx_MsgPair 	m;
					if (tiGtTx_getMsgPair( &m , &s ) == 0)
						break;
					if (tiGtTx_msgConv_num >= size) {
						size 			  = size + size / 2;
						tiGtTx_msgConvTbl = (TiGtTx_MsgPair*) REALLOC(tiGtTx_msgConvTbl, (sizeof tiGtTx_msgConvTbl[0]) * size );
						if (tiGtTx_msgConvTbl == NULL) {
							break;
						}
					}
					tiGtTx_msgConvTbl_insert(tiGtTx_msgConvTbl, (int*)&tiGtTx_msgConv_num, &m);
				}

				FREE( tiGtTx_workBuf );
				tiGtTx_workBuf = NULL;
			}
		}
		FREE(msgBuf);
	}
	tiGtTx_domainName		= "";
	tiGtTx_file_name		= "";
	return domainName;
}



#if 1
void tiny_gettext_term(void)
{
	TiGtTx_MsgPair*	t = tiGtTx_msgConvTbl;
	if (t) {
		unsigned		i;
		for (i = 0; i < tiGtTx_msgConv_num; ++i) {
			if (t->id_)
				FREE(t->id_);
			if (t->str_)
				FREE(t->str_);
			++t;
		}
		FREE(tiGtTx_msgConvTbl);
		tiGtTx_msgConvTbl = NULL;
	}
}
#endif



/* -------------------------------------------------------------------------*/

static char* tiGtTx_getMsgFileName(void)
{
	char *	 basePath	= tiGtTx_getBasePath();
	unsigned baseLen	= strlen(basePath);
	char *	 msgName	= 0;
	if (basePath[baseLen] == '\0') {
		char *	 localeStr	  = tiGtTx_getLocaleStr();
		unsigned localeStrLen = strlen(localeStr);
		unsigned msgNameLen   = baseLen + 1 + localeStrLen + 4;		// baseName + "."+ lang_str + ".po\0"
		assert(msgNameLen <= TIGTTX_MAX_PATH);
		msgName = (char*)CALLOC(1, msgNameLen + 16);
		if (localeStr[0]) {
			FILE *	fp;
			sprintf( msgName, "%s.%s.po", basePath, localeStr );
			// exist msgName ?
			fp = fopen(msgName,"rb");
			if (fp)
				fclose(fp);
			else
				localeStr = 0;
		}
		if (localeStr == 0 || localeStr[0] == 0) {
			sprintf( msgName, "%s.po", basePath );
		}
	}
	if (basePath[0])	// * when [0] == '\0', not allocate.
		FREE(basePath);
	return msgName;
}



static char* tiGtTx_getLocaleStr(void)
{
  #ifdef _WIN32
	static char buf[22];
	if (buf[0] == 0) {
		unsigned 	locale_id = GetThreadLocale();
		sprintf(buf, "%d", locale_id );
	}
	return buf;
  #else
	return getenv("LANG");
  #endif
}



static char* tiGtTx_getBasePath(void)
{
 #ifdef _WIN32
	char* 	name = (char*)CALLOC(1, TIGTTX_MAX_PATH + 5);
	char* 	p;
	if (name == NULL)
		return "";

	//strncpy(name, _pgmptr, TIGTTX_MAX_PATH );	// defined _MSC_VER
	GetModuleFileName(NULL, name, TIGTTX_MAX_PATH );

	p = strrchr((char *) name, '.');
	if (p == NULL || strlen(p) == 0) {
		FREE( name );
		return "";
	}
	*p = '\0';
	return name;
 #else	/* others( linux,unix ) */
	char*		name;
	int			l;
	const char* home = getenv("HOME");
	if (home == 0 || home[0] == 0) {
		home = ".";
	}
	l	 = strlen(home) + 2 + strlen(tiGtTx_domainName) + 31;
	name = CALLOC( 1, l + 1 );
	if (name == NULL)
		return "";

	sprintf(name, "%s/.%s", home, tiGtTx_domainName);
	return name;
 #endif
}



/* -------------------------------------------------------------------------*/

static char* tiGtTx_fileLoad(const char* fname)
{
	unsigned	size;
	char*		buf;
	FILE* 		fp;
	size_t		l	= 0;

	if (fname == 0 || *fname == 0)
		return NULL;

	fp = fopen(fname, "rb");
	if (fp == NULL)
		return NULL;

  #if 1
	fseek(fp, 0, 2);
	size = ftell(fp);
	rewind(fp);
  #else
	{
		struct stat fl_stat;
		fstat(fileno(fp), &fl_stat);
		size	= fl_stat.st_size;
	}
  #endif

	if (size > TIGTTX_FILE_SIZE_MAX) {
		fclose(fp);
		TIGTTX_ERRMSG((STDERR, "%s : file size is too large.\n", fname));
		return NULL;
	}
	buf 	= (char*)CALLOC( 1, size + 16 );		// +16 ... for '\0', for strcmp(p,"msgid") check.
	if (buf)
		l = fread(buf, 1, size, fp);
	fclose(fp);

	if (l != size || size == 0) {
		FREE( buf );
		buf = NULL;
	} else {
		buf[size] = '\0';
	}

	return buf;
}



// -----------------------------------------------------------

static int tiGtTx_getMsgPair(TiGtTx_MsgPair* pMsgPair, char const* * ppStr)
{
	const char* s	 = (const char *) *ppStr;
	int   		mode = 0;

	pMsgPair->id_	 = 0;
	pMsgPair->str_	 = 0;

	for (;;) {
		int c;
		s = tiGtTx_skipspc(s);
		c = *s;
		if (c == 0)
			break;

		if (c == 'm') {
			if (memcmp("msgid", s, 5) == 0) {
				if (mode == 1) {
					TIGTTX_ERRMSG((STDERR, "%s (%d) : no msgstr\n", tiGtTx_file_name, tiGtTx_file_line));
					if (pMsgPair->id_)
						FREE( pMsgPair->id_ );
					pMsgPair->id_ = 0;
				}
				s	 = tiGtTx_skipspc(s + 5);
				if (*s == '"') {
					pMsgPair->id_ = tiGtTx_getStr(&s);
					if (pMsgPair->id_)
						mode = 1;
					continue;
				}

			} else if (memcmp("msgstr", s, 6) == 0) {
				int f = 0;
				s += 6;
				if (*s == '[') {
					const char * p = strchr(s+1, ']');
					if (p)
						s = p+1;
					f = 1;
				}
				s	 = tiGtTx_skipspc(s);
				if (*s == '"') {
					pMsgPair->str_ = tiGtTx_getStr(&s);
					if (mode && pMsgPair->str_)
						break;
					if (f == 0)
						TIGTTX_ERRMSG((STDERR, "%s (%d) : no msgid\n", tiGtTx_file_name, tiGtTx_file_line));
					if (pMsgPair->str_)
						FREE( pMsgPair->str_ );
					pMsgPair->str_ = NULL;
					continue;
				}
			}
		}

		TIGTTX_ERRMSG((STDERR, "%s (%d) : bad character.(%#x)\n", tiGtTx_file_name, tiGtTx_file_line, c));
		do {
			c = *++s;
		} while (c != 0 && c != '\n' && c != '\r');
	}

	*ppStr = s;
	return pMsgPair->str_ != NULL;
}



static char* tiGtTx_getStr(char const* * ppSrc)
{
	char *		d	= tiGtTx_workBuf;
	const char* s	= *ppSrc;
	size_t		l;

	for (;;) {
		s = tiGtTx_skipspc(s);
		if (*s == '\0')
			break;
		if (*s != '"')
			break; //x return NULL;
		++s;
		for (;;) {
			char c = *s++;
			if (c == '\0')
				return NULL;

			if (c == '"')
				break;

			if (c == '\\') {
				tiGtTx_strEsc(&d, &s);
			} else {
				*d++ = c;
				if (IsMbbLead((unsigned char)c) && *s)
					*d++ = *s++;
			}
		}
	}
	*d	   	= 0;
	*ppSrc 	= s;
	l		= d - tiGtTx_workBuf;
	d		= (char*) CALLOC(1, l+1);
	if (d)
		memcpy(d, tiGtTx_workBuf, l);
	return d;
}



static void tiGtTx_strEsc(char** ppDst, char const* * ppSrc)
{
	const char* s = *ppSrc;
	char* 		d = *ppDst;
	int			k;
	int 		c = *s++;
	switch (c) {
	case 0: 	return;
	case 'n':	*d++ = '\n'; break;
	case 'r':	*d++ = '\r'; break;
	case 'a':	*d++ = '\a'; break;
	case 'b':	*d++ = '\b'; break;
	case 'f':	*d++ = '\f'; break;
	case 'v':	*d++ = '\v'; break;
	case 't':	*d++ = '\t'; break;
	case 'x':
		k = 1;
		goto XDIGIT;

	case 'u':
		k = 2;
		goto XDIGIT;

	case 'U':
		k = 4;

  XDIGIT:
  		do {
			int val = 0;
			int l	= 2;
			do {
				c = *s;
				if ('0' <= c && c <= '9') {
					c -= '0';
				} else if ('A' <= c && c <= 'F') {
					c  = c - 'A' + 10;
					val = (val << 4) | (c - 'A' + 10);
				} else if ('a' <= c && c <= 'f') {
					c  = c - 'a' + 10;
				} else {
					k = 1;
					break;
				}
				++s;
				val = (val << 4) | (c);
			} while (--l);
			*d++ = (char)val;
		} while (--k);
		break;

	default:
		if (c >= '0' && c <= '7') {
			int val = c - '0';
			int l	= 3-1;
			do {
				c = *s;
				c -= '0';
				if (c < 0 || 7 < c) {
					break;
				}
				++s;
				val = (val << 3) | (c);
			} while (--l);
			*d++ = (char)val;
		} else {
			*d++ = (char)c;
		}
	}

	*ppDst = d;
	*ppSrc = s;
}



static const char* tiGtTx_skipspc(const char* str)
{
	const unsigned char* s = (const unsigned char*)str;
	int   c;
	do {
		c = *s++;
		if (c == '\n') {
			++ tiGtTx_file_line;
		} else if (c == '\r') {
			if (*s == '\n')
				++s;
			++ tiGtTx_file_line;
		} else if (c == '#') {
			do {
				c = *s++;
			} while (c != 0 && c != '\n' && c != '\r');
			--s;
		}
	} while (c && isspace(c));
	if (*s)
		--s;
	return (const char*)s;
}



/* -------------------------------------------------------------------------*/

static int tiGtTx_msgConvTbl_insert(TiGtTx_MsgPair tbl[], int* tblcnt, const TiGtTx_MsgPair* key)
{
	int  low, mid, f, hi;

	hi	= *tblcnt;
	mid = low = 0;
	while (low < hi) {
		mid = (low + hi - 1) / 2;
		f = strcmp(key->id_, tbl[mid].id_);
		if (f < 0) {
			hi = mid;
		} else if (f > 0) {
			mid++;
			low = mid;
		} else {
			assert(0);
			return -1;
		}
	}
	++(*tblcnt);
	for (hi = *tblcnt; --hi > mid;) {
		tbl[hi] = tbl[hi - 1];
	}
	tbl[mid] = *key;
	return 0;
}



static int tiGtTx_msgConvTbl_search(TiGtTx_MsgPair tbl[], int nn, const char* key)
{
	int 	low, mid, f;

	low = 0;
	while (low < nn) {
		mid = (low + nn - 1) / 2;
		f = strcmp(key, tbl[mid].id_);
		if (f < 0)
			nn = mid;
		else if (f > 0)
			low = mid + 1;
		else
			return mid;
	}
	return -1;
}



/* -------------------------------------------------------------------------*/
